"""
molecularprofiles

The package provides a set of useful routines
for the retrieval and analysis of the meteorological data
from different data assimilation systems
"""

from .version import __version__
